{
    "flavors": [
        {
            "id": "1",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/1",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/1",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.tiny",
            "description": null
        },
        {
            "id": "2",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/2",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/2",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.small",
            "description": null
        },
        {
            "id": "3",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/3",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/3",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.medium",
            "description": null
        },
        {
            "id": "4",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/4",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/4",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.large",
            "description": null
        },
        {
            "id": "5",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/5",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/5",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.xlarge",
            "description": null
        },
        {
            "id": "6",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/6",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/6",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.tiny.specs",
            "description": null
        },
        {
            "id": "%(flavorid)s",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/flavors/%(flavorid)s",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/flavors/%(flavorid)s",
                    "rel": "bookmark"
                }
            ],
            "name": "m1.small.description",
            "description": "test description"
        }
    ]
}
