{
    "snapshots": [
        {
            "createdAt": "%(strtime)s",
            "displayDescription": "Default description",
            "displayName": "Default name",
            "id": 100,
            "size": 100,
            "status": "available",
            "volumeId": 12
        },
        {
            "createdAt": "%(strtime)s",
            "displayDescription": "Default description",
            "displayName": "Default name",
            "id": 101,
            "size": 100,
            "status": "available",
            "volumeId": 12
        },
        {
            "createdAt": "%(strtime)s",
            "displayDescription": "Default description",
            "displayName": "Default name",
            "id": 102,
            "size": 100,
            "status": "available",
            "volumeId": 12
        }
    ]
}
