{
    "servers": [
        {
            "id": "%(uuid)s",
            "status": "UNKNOWN",
            "links": [
                {
                    "href": "%(versioned_compute_endpoint)s/servers/%(uuid)s",
                    "rel": "self"
                },
                {
                    "href": "%(compute_endpoint)s/servers/%(uuid)s",
                    "rel": "bookmark"
                }
            ]
        }
    ]
}