{
    "aggregate": {
        "availability_zone": "london",
        "created_at": "%(strtime)s",
        "deleted": false,
        "deleted_at": null,
        "hosts": [
            "%(compute_host)s"
        ],
        "id": 1,
        "metadata": {
            "availability_zone": "london"
        },
        "name": "name",
        "updated_at": null
    }
}
