{
    "flavor": {
        "name": "%(flavor_name)s",
        "ram": 1024,
        "vcpus": 2,
        "disk": 10,
        "id": "%(flavor_id)s",
        "rxtx_factor": 2.0,
        "description": "test description"
    }
}
