{
    "server_groups": [
        {
            "id": "%(id)s",
            "name": "test",
            "policies": ["anti-affinity"],
            "members": [],
            "metadata": {}
        }
   ]
}
