{
    "keypair": {
        "public_key": "%(public_key)s",
        "name": "%(keypair_name)s",
        "type": "%(keypair_type)s",
        "fingerprint": "%(fingerprint)s",
        "user_id": "fake",
        "deleted": false,
        "created_at": "%(strtime)s",
        "updated_at": null,
        "deleted_at": null,
        "id": 1
    }
}
