{
    "snapshot": {
        "createdAt": "%(strtime)s",
        "displayDescription": "%(description)s",
        "displayName": "%(snapshot_name)s",
        "id": 100,
        "size": 100,
        "status": "available",
        "volumeId": "%(uuid)s"
    }
}
