{
    "rebuild" : {
        "imageRef" : "%(uuid)s",
        "accessIPv4" : "%(access_ip_v4)s",
        "accessIPv6" : "%(access_ip_v6)s",
        "adminPass" : "%(pass)s",
        "metadata" : {
            "meta_var" : "meta_val"
        },
        "name" : "%(name)s",
        "OS-DCF:diskConfig": "%(disk_config)s",
        "personality" : [
            {
                "path" : "/etc/banner.txt",
                "contents" : "ICAgICAgDQoiQSBjbG91ZCBkb2VzIG5vdCBrbm93IHdoeSBp dCBtb3ZlcyBpbiBqdXN0IHN1Y2ggYSBkaXJlY3Rpb24gYW5k IGF0IHN1Y2ggYSBzcGVlZC4uLkl0IGZlZWxzIGFuIGltcHVs c2lvbi4uLnRoaXMgaXMgdGhlIHBsYWNlIHRvIGdvIG5vdy4g QnV0IHRoZSBza3kga25vd3MgdGhlIHJlYXNvbnMgYW5kIHRo ZSBwYXR0ZXJucyBiZWhpbmQgYWxsIGNsb3VkcywgYW5kIHlv dSB3aWxsIGtub3csIHRvbywgd2hlbiB5b3UgbGlmdCB5b3Vy c2VsZiBoaWdoIGVub3VnaCB0byBzZWUgYmV5b25kIGhvcml6 b25zLiINCg0KLVJpY2hhcmQgQmFjaA=="
            }
        ],
        "preserve_ephemeral": %(preserve_ephemeral)s,
        "description" : "%(description)s"
    }
}
