{
    "volumeAttachment": {
        "volumeId": "%(volume_id)s",
        "id": "%(volume_id)s",
        "serverId": "%(server_id)s",
        "device": "%(device)s",
        "tag": "%(tag)s",
        "delete_on_termination": true
    }
}
